/*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jsvg;

import com.jsevy.jxml.*;

import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.io.StringWriter;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;


/**
 * Class representing a Scalable Vector Graphics document, which owns an svgGraphics on which drawing commands can be made. The document's
 * toSVGString method can then be called to generate the SVG text corresponding to the drawing done on the Graphics object.
 * 
 * The typical workflow is as follows:
 * 
 * <pre>
 * 
 * // Create an svg document and get its associated SVGGraphics instance
 * SVGDocument svgDocument = new SVGDocument(); 
 * SVGGraphics svgGraphics = svgDocument.getGraphics(); 
 * 
 * // Do drawing commands as on any other Graphics. If you have a paint(Graphics) method, 
 * // you can just use it with the SVGGraphics instance since it's a subclass of Graphics. 
 * graphics.setColor(Color.RED);  
 * graphics.setStroke(new BasicStroke(3));
 * graphics.drawLine(0, 0, 1000, 500); 
 * graphics.drawRect(1000, 500, 150, 150); 
 * graphics.drawRoundRect(20, 200, 130, 100, 20, 10); 
 * 
 * // Get the flat SVG output as a string - it's just XML text - and  save  in a file for use 
 * // with LibreOffice Draw or any other draw package that accepts flat SVG (.svg) files
 * String stringOutput = svgDocument.toSVGString(); 
 * String filePath = "path/to/file.svg"; 
 * FileWriter fileWriter = new FileWriter(filePath); 
 * fileWriter.write(svgText); 
 * fileWriter.flush(); 
 * fileWriter.close();
 * 
 * </pre>
 * 
 * @author jsevy
 *
 */
public class SVGDocument extends XMLDocument
{
    
    //private Vector<SVGEntity> entities;
    
    private SVGGraphics graphics;
    
    private SVGStyles styles;
    
    
    private Rectangle2D.Double documentBounds = new Rectangle2D.Double(0, 0, 10, 10);

    private boolean autosize = true;
    
    public static String svgNamespaceURI = "http://www.w3.org/2000/svg";
    
    
    /**
     * Create a new SVGDocument
     */
    public SVGDocument()
    {
        super("svg", svgNamespaceURI);
        
        this.graphics = new SVGGraphics(this);
        
        //this.entities = new Vector<SVGEntity>();
        this.styles = new SVGStyles();
        
        populateXML();
    }
    
    
    
    
    /** Set the units for interpreting the dimension values in the SVG file.
     *
     *  @param unitsCode  SVG code indicating units:
     *      0 = Unitless
     *      1 = Inches
     *      2 = Feet
     *      3 = Miles
     *      4 = Millimeter
     *      5 = Centimeters
     *      6 = Meters
     *      7 = Kilometers
     *      8 = Microinches
     *      9 = Mils
     *      10 = Yards
     * 
     */
    public void setUnits(int unitsCode)
    {
        String[] unitsMatrix = 
            {
            /* 0 = */ "",
            /* 1 = */ "in",
            /* 2 = */ "ft",
            /* 3 = */ "mi",
            /* 4 = */ "mm",
            /* 5 = */ "cm",
            /* 6 = */ "m",
            /* 7 = */ "km",
            /* 8 = */ "uin",
            /* 9 = */ "mils",
            /* 10 = */ "yd",
        };
        
        SVGEntity.units = unitsMatrix[unitsCode];
    }
    
    
    
    /**
     * Set the document width and height, which equates to the overall viewBox width and height for the SVG 
     * drawing added as an attribute to the top-level svg element. Note that the units used are those set with a 
     * call to setUnits(). This also turns off autosizing of the document.
     * 
     * @param documentWidth    Document/viewBox width, in units set by call to setUnits()
     * @param documentHeight   Document/viewBox height, in units set by call to setUnits()
     */
    public void setDocumentSize(double documentWidth, double documentHeight)
    {
        this.documentBounds.width = documentWidth;
        this.documentBounds.height = documentHeight;
        
        this.autosize = false;
    }
    
    
    /**
     * Set the document rectangle, which equates to the viewBox for the SVG drawing which
     * is added as an attribute to the top-level svg element. Note that the units used are those set 
     * with a call to setUnits(). This also turns off autosizing of the document.
     * 
     * @param bounds    Document/viewBox rectangle, in units set by call to setUnits()
     */
    public void setDocumentBounds(Rectangle2D.Double bounds)
    {
        this.documentBounds = bounds;
        
        this.autosize = false;
    }
    
    
    /**
     * Return a copy of the rectangle defining the viewport of the document.
     * 
     * @return  A copy of the rectangle defining the document viewport
     */
    public Rectangle2D.Double getDocumentBounds()
    {
        return new Rectangle2D.Double(documentBounds.x, documentBounds.y, documentBounds.width, documentBounds.height);
    }
    
    
    /**
     * Specify whether or not to autosize the document, which will set the viewport to the minimum size needed to
     * enclose all of the graphics entities. The default is true, but this may be set to false by a call to
     * this method, at which point a default viewport will be used, or by a call to either the setDocumentSize() or 
     * setDocumentBounds() methods, which explicitly specify the document (viewport) size.
     * 
     * @param autosize    Boolean specifying whether the document should be automatically sized to enclose all
     *                    of the graphical entities in the generated SVG; default is true
     */
    public void autosizeDocument(boolean autosize)
    {
        this.autosize  = autosize;
    }
    
    
    
    /**
     * Return the smallest rectangle which encloses the graphical elements included in the drawing up to the point of this call.
     * This can be used to specify a viewport for the SVG, for example if it is desired to include margins around the 
     * graphics, by calling after drawing is completed and using the result to define an enlarged rectangle for use in 
     * a call to setDocumentBounds().
     * 
     * @return  Rectangle enclosing all of the graphical elements in the design at the point of the call to this method
     */
    public Rectangle2D.Double getEntityBounds()
    {
        Rectangle2D.Double bounds = new Rectangle2D.Double();
        
        // run through entities to find max and min
        XMLElement documentElement = (XMLElement)this.getDocumentElement();
        XMLNodeList svgEntityList = (XMLNodeList)documentElement.getChildNodes();
        if (svgEntityList.size() > 0)
        {
            SVGEntity entity = (SVGEntity)svgEntityList.elementAt(0);
            Rectangle2D.Double entityBounds = (Rectangle2D.Double)entity.getBounds();
            double xMin = entityBounds.x;
            double yMin = entityBounds.y;
            double xMax = entityBounds.x + entityBounds.width;
            double yMax = entityBounds.y + entityBounds.height;
            
            for (int i = 1; i < svgEntityList.size(); i++)
            {
                entity = (SVGEntity)svgEntityList.elementAt(i);
                entityBounds = (Rectangle2D.Double)entity.getBounds();

                xMin = Math.min(xMin, entityBounds.x);
                yMin = Math.min(yMin, entityBounds.y);
                xMax = Math.max(xMax, entityBounds.x + entityBounds.width);
                yMax = Math.max(yMax, entityBounds.y + entityBounds.height);
                
            }
            
            bounds.x = xMin;
            bounds.y = yMin;
            bounds.width = xMax - xMin;
            bounds.height = yMax - yMin;
        }
        
        return bounds;
    }
    
    
    
    /**
     * Set the scale factor for dimensions in the Java drawing - how many Java units/pixels
     * correspond to one unit in the SVG output. Note that the units can be set with a call to 
     * setUnits
     * 
     * @param javaUnitsPerUnit  How many Java units/pixels correspond to one SVG unit
     */
    public void setScaleFactor(double javaUnitsPerUnit)
    {
        SVGEntity.scaleFactor = javaUnitsPerUnit;
    }
    
    
    
    
    /**
     * Get the SVGGraphics associated with this document for use with standard Graphics drawing operations to generate
     * an svg text representation.
     * 
     * @return	The SVGGraphics associated with this document, on which standard Java Graphics drawing calls can be made
     */
    public SVGGraphics getGraphics()
    {
        return this.graphics;
    }
    
    
    /**
     * Get a graphics style corresponding to the current Graphics2D instance, and whether the shape
     * should be stroked and/or filled. Will return an existing corresponding style if there is one,
     * or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @param stroked   True if shape is to be stroked with current color of supplied graphics instance
     * @param filled    True if shape is to be filled with current color of supplied graphics instance
     * @return          Corresponding SVGGraphicsStyle
     */
    protected SVGGraphicsStyle getGraphicsStyle(Graphics2D graphics, boolean stroked, boolean filled)
    {
        return styles.getGraphicsStyle(graphics, stroked, filled);
    }
    
    
    /**
     * Get a text style corresponding to the current Graphics2D instance. Will return an existing corresponding
     * style if there is one, or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @return          Corresponding SVGTextStyle
     */
    protected SVGTextStyle getTextStyle(Graphics2D graphics)
    {
        return styles.getTextStyle(graphics);
    }
    
    
    private void populateXML()
    {
        // namespace attribute added in document creation
        //this.getDocumentElement().setAttribute("xmlns:", svgNamespaceURI);
    }
    
    
    
    /**
     * Return the SVG text associated with this SVG document. This includes the header and
     * entities populated with content generated by graphics calls on the associated SVGGraphics.
     * 
     * @return The SVG text associated with this document.
     * @throws TransformerException 
     */
    public String toSVGString()
    {
        // first, add viewbox to set user scale; wait until XML generation so the units and page size can be set if desired
        if (autosize) 
        {
            // set the document size to enclose all graphics elements
            Rectangle2D.Double bounds = this.getEntityBounds();
            this.setDocumentBounds(bounds);
        }
        
        this.getDocumentElement().setAttribute("width", SVGObject.addUnits(documentBounds.width));
        this.getDocumentElement().setAttribute("height", SVGObject.addUnits(documentBounds.height));
        this.getDocumentElement().setAttribute("viewBox", documentBounds.x + " " + documentBounds.y +" " + documentBounds.width + " " + documentBounds.height);
        
        DOMSource domSource = new DOMSource(this);
        StringWriter stringWriter = new StringWriter();
        
        try
        {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            
            transformer.transform(domSource, new StreamResult(stringWriter));
        }
        catch (TransformerException e)
        {
            // really shouldn't ever happen...
            System.out.println("TransformerException while creating SVG XML: " + e.getMessage());
        }
        
        //xmlBuffer.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + "\n");
        //xmlBuffer.append(this.toXMLString(0, 2));
        
        StringBuffer xmlBuffer = stringWriter.getBuffer();
        return xmlBuffer.toString();
    }
    
    
    /**
     * Utility method used by the associated SVGGraphics object.
     * 
     * @param entity an svgEntity instance
     */
    public void addEntity(SVGEntity entity)
    {
        //entities.add(entity);
        this.getDocumentElement().appendChild(entity);
    }
    
}